/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.nlpcraft.internal.intent.compiler.functions

import org.apache.nlpcraft.*
import internal.intent.*
import internal.intent.compiler.*
import internal.intent.compiler.functions.*
import nlp.util.*
import org.scalatest.funsuite.AnyFunSuite

import java.util
import java.util.UUID
import scala.language.implicitConversions
/**
  *
  */
private[functions] object NCIDLFunctions:
    private final val MODEL_ID = "test.mdl.id"

    given Conversion[NCTestToken, Seq[NCTestToken]] with
        def apply(t: NCTestToken): Seq[NCTestToken] = Seq(t)

    given Conversion[String, TestDesc] with
        def apply(truth: String): TestDesc = TestDesc(truth)

    /**
      *
      * @param truth
      * @param entity
      * @param idlCtx
      * @param expectedRes
      * @param entitiesUsed
      */
    case class TestDesc(
        truth: String,
        entity: Option[NCEntity] = None,
        idlCtx: NCIDLContext,
        expectedRes: Boolean = true,
        entitiesUsed: Option[Int] = None
    ):
        lazy val term: NCIDLTerm =
            val intents = new NCIDLCompiler(idlCtx.mdlCfg).compile(s"intent=i term(t)={$truth}", MODEL_ID)

            require(intents.size == 1)
            require(intents.head.terms.sizeIs == 1)

            intents.head.terms.head

        override def toString: String =
            entity match
                case Some(e) => s"Predicate [body='$truth', entity=${e2s(e)}]"
                case None => s"Predicate '$truth'"

    object TestDesc:
        /**
          *
          * @param truth
          */
        def apply(truth: String): TestDesc =
            new TestDesc(truth = truth, idlCtx = mkIdlContext())

        /**
          *
          * @param truth
          * @param entity
          * @param idlCtx
          */
        def apply(truth: String, entity: NCEntity, idlCtx: NCIDLContext): TestDesc =
            new TestDesc(truth = truth, entity = entity.?, idlCtx = idlCtx)

        /**
          *
          * @param truth
          * @param entity
          */
        def apply(truth: String, entity: NCEntity): TestDesc =
            new TestDesc(truth = truth, entity = entity.?, idlCtx = mkIdlContext(entities = Seq(entity)))



    /**
      *
      * @param e
      */
    private def e2s(e: NCEntity): String = s"${e.getType} (${e.getTokens.map(_.getText).mkString(" ")})"

    /**
      *
      * @param entities
      * @param cfg
      * @param reqId
      * @param txt
      * @param ts
      * @param userId
      * @param reqData
      * @param intentMeta
      * @param convMeta
      * @param fragMeta
      */
    def mkIdlContext(
        entities: Seq[NCEntity] = Seq.empty,
        cfg: NCModelConfig = CFG,
        reqId: String = null,
        txt: String = null,
        ts: Long = 0,
        userId: String = null,
        reqData: Map[String, AnyRef] = Map.empty,
        intentMeta: Map[String, AnyRef] = Map.empty,
        convMeta: Map[String, AnyRef] = Map.empty,
        fragMeta: Map[String, AnyRef] = Map.empty
    ): NCIDLContext =
        NCIDLContext(
            cfg,
            entities,
            intentMeta = intentMeta,
            convMeta = convMeta,
            fragMeta = fragMeta,
            req = NCTestRequest(txt = txt, userId = userId, reqId = reqId, ts = ts, data = reqData)
        )

    /**
      *
      * @param typ
      * @param reqId
      * @param value
      * @param groups
      * @param meta
      * @param tokens
      */
    def mkEntity(
        typ: String = UUID.randomUUID().toString,
        reqId: String = UUID.randomUUID().toString,
        value: String = null,
        groups: Set[String] = null,
        meta: Map[String, AnyRef] = Map.empty[String, AnyRef],
        tokens: Seq[NCTestToken]
    ): NCEntity =
        require(tokens.nonEmpty)

        NCTestEntity(typ, reqId, groups, meta, tokens)

import org.apache.nlpcraft.internal.intent.compiler.functions.NCIDLFunctions.*

/**
  * Tests for IDL functions.
  */
private[functions] trait NCIDLFunctions extends AnyFunSuite:
    /**
      *
      * @param funcs
      */
    protected def test(funcs: TestDesc*): Unit =
        for (f <- funcs)
            val item =
                try
                    // Process declarations.
                    f.idlCtx.vars ++= f.term.decls

                    // Execute term's predicate.
                    val ent = f.entity.orNull
                    val idEnt =
                        if ent != null then
                            require(ent.getTokens != null && ent.getTokens.nonEmpty)

                            NCIDLEntity(ent, ent.getTokens.minBy(_.getIndex).getIndex)
                        else
                            null

                    f.term.pred.apply(idEnt, f.idlCtx)
                catch
                    case e: NCException => throw e
                    case e: Exception => throw new Exception(s"Execution error processing: $f", e)

            item.value match
                case b: java.lang.Boolean => require(if f.expectedRes then b else !b, s"Unexpected '$b' result for: $f")
                case _ =>
                    require(
                        requirement = false,
                        s"Unexpected result type [resType=${
                            if item.value == null then "null"
                            else item.value.getClass.getName
                        }, resValue=${item.value}, function=$f]"
                    )

            f.entitiesUsed match
                case Some(exp) =>
                    require(
                        exp == item.entUse,
                        s"Unexpected tokens used [expectedTokensUsed=$exp, resultEntityUsed=${item.entUse}, function=$f]"
                    )

                case None => // No-op.

    /**
      *
      * @param funcs
      */
    protected def expectError(funcs: TestDesc*): Unit =
        for (f <- funcs)
            try
                test(f)

                require(false)
            catch
                case e: Exception =>
                    println(s"Expected error: ${e.getLocalizedMessage}")

                    var cause = e.getCause

                    while (cause != null)
                        println(s"  Cause: ${cause.getLocalizedMessage} (${cause.getClass.getName})")

                        cause = cause.getCause